# -----------------------------------------------------------------------------------------------------------------
# This script calibrates the TreeAge Pro model MarkovWeibull.trex
#  Specifically, calibration adjusts 6 parameters for 3 Weibull transitions to best fit "observed results".
#
#  Sets initial test parameters for the model.
#  Configures the Nelder-Mead algorithm for optimization with:
#   - a function to analyze the TP model and gather specific results for comparison to "observed results"
#  Executes the algorithm, which repeatedly analyzes the model providing new parameters while attempting to
#   minimize the difference between calculated results and "observed results"
#  Final parameters from algorithm are returned as calibrated parameters.
#
# To execute the Python script:
#  Need to have a Python IDE installed and configured with NumPy, SciPy and COMTYPES libraries.
#   Run this command in the IDE - execfile("C:/Python27/Scripts/Calibration with Nelder-Mead.py")
#
# Calibration adjusts 6 parameters for 3 Weibull transitions using example MarkovWeibull.trex model.
# -----------------------------------------------------------------------------------------------------------------

# COMTYPES, SCIPY and NUMPY libraries are required for this script to run.
# Import necessary components:

from comtypes.client import CreateObject
from datetime import datetime
from scipy.optimize import minimize
from scipy.optimize import basinhopping
import numpy as np


# Function which sets a variable within TreeAgePro model to a specified value
def set_tree_variable(variable, value) :
    tree.defineRootVariable(variable, str(value))

# Initialize Object Interface instance to access TreeAge Pro functionality
treeage = CreateObject("TreeAgeProOI.ApplicationObj")

# Open specific TreeAge Pro model.  Adjust the path and name specific to your computer environment.
tree = treeage.getTree("C:/Python27/Scripts/Models/MarkovWeibull.trex")
# For Rankings and microsimulation analysis it is sufficient to set the root of the TreeAge Pro model to be selected.
markov_root = tree.getRoot()

# If cohort analysis is required then the root of the analysis has to be a Markov node.
# The following code is an example that selects top most child node of the root node.
# If the topmost child BranchIndex(0) is indeed a Markov node the cohort analysis can be run.
# If the node is not a Markov node the cohort analysis will fail and script will break on trying to read the empty report.
# Some additional error checking could be added for that case.
#
#markov_root = tree.getRoot().getBranchByIndex(0)

# Calibration Data.  The desired proportions of the cohort at 3 different times.
# Specification of times when the observed cohort proportions where gathered.
CalibrationPoints = np.array([1.0, 2.0, 3.0])
# Specication of cohort proportions [State1, State2, ... StateN] for each observed time point.
CalibrationValues = np.array([[0.66834376, 0.22657988, 0.10507636],
                             [0.24700605, 0.44672552, 0.30626843],
                             [0.05377982, 0.40597475, 0.54024543]])

# Set the values of the initial values for rates and shape parameters.
# The starting points for the optimization algorithm to find better parameter estimates.

initial_rate12 = 0.29
initial_shape12 = 1.85
initial_rate13 = 0.09
initial_shape13 = 1.35
initial_rate23 = 0.19
initial_shape23 = 1.45

# Pack parameters into a single array as required by the optimization function.
initial_params = [initial_rate12, initial_shape12,
                  initial_rate13, initial_shape13,
                  initial_rate23, initial_shape23]

# Set other parameters for the TreeAge Pro analysis.
set_tree_variable("TimeHorizon", 4)
set_tree_variable("vPeriodsPerYear", 12)

# Function which sets the values of all parameters within TreeAge Pro
def set_rates(rates) :
#    print "current rates: %s" % str(rates)
    set_tree_variable("vRate_g_12", rates[0])
    set_tree_variable("vShape_k_12", rates[1])
    set_tree_variable("vRate_g_13", rates[2])
    set_tree_variable("vShape_k_13", rates[3])
    set_tree_variable("vRate_g_23", rates[4])
    set_tree_variable("vShape_k_23", rates[5])

    
# calculate goodness of fit from an analysis results
# refer to print_headers(report) function below for column index lookup.
#
# The data in reports is organized in rows and columns.
# You may need to experiment to find which row is needed to access the data of interest.
#
# Run the analysis once without invoking the optimization function and then experiment
# by printing selected rows.  In particular the cohort reports contains multiple rows and
# the totals are not contained in the last few rows, so printing several of the last rows
# will allow you to find the data of interest.
#
# Summary type reports (e.g. rankings) will contain mean values in row 1.

def calc_goodness(report) :
    # retrieve the headers from the report returned by TreeAge Pro analysis
    headers = report.getTextReport().getHeaders()
    # locate state membership data from the payoffs
    # SXTY ->  State X membership proportion at Time Y
    
    S1T1 = headers.index("payoff4")
    S2T1 = headers.index("payoff5")
    S3T1 = headers.index("payoff6")
    S1T2 = headers.index("payoff7")
    S2T2 = headers.index("payoff8")
    S3T2 = headers.index("payoff9")
    S1T3 = headers.index("payoff10")
    S2T3 = headers.index("payoff11")
    S3T3 = headers.index("payoff12")
    
    rows = report.getTextReport().getRows()
# Calculate the sum of squares of differences between simulated proportions and desired proportions by state and time.
    result = 0
    result = result + np.power(float(rows[1][S1T1]) - CalibrationValues[0,0],2)
    result = result + np.power(float(rows[1][S2T1]) - CalibrationValues[0,1],2)
    result = result + np.power(float(rows[1][S3T1]) - CalibrationValues[0,2],2)
    result = result + np.power(float(rows[1][S1T2]) - CalibrationValues[1,0],2)
    result = result + np.power(float(rows[1][S2T2]) - CalibrationValues[1,1],2)
    result = result + np.power(float(rows[1][S3T2]) - CalibrationValues[1,2],2)
    result = result + np.power(float(rows[1][S1T3]) - CalibrationValues[2,0],2)
    result = result + np.power(float(rows[1][S2T3]) - CalibrationValues[2,1],2)
    result = result + np.power(float(rows[1][S3T3]) - CalibrationValues[2,2],2)
    return result

# Just what is says it invokes specified analysis on the TreeAge Pro model and returns analysis report object.
def run_analysis() :
    params = CreateObject("TreeAgeProOI.ParamsObj")
    report = tree.runAnalysis("rankings", params, markov_root)
    return report

# This function writes the headers from the analysis report to a text file "headers.txt"
# Use this file to lookup the particular column index to the column of interest.
def print_headers(report) :
    file = open("headers.txt", "w")
    file.write(str(report.getHeaders()))
    file.close()

# This is the main function that sets up parameters and runs the analysis.
# This function is actually called from within the calibration routine. 
def markov_model (input) :
    set_rates(input)
    report = run_analysis()
# Print headers only if you are changing model structure or type of report and you need to located you payoffs of interest.
#    print_headers(report.getTextReport())
#
#   Calculate and return the Goodness of Fit value.
    return calc_goodness(report)

print("Start Calibration...")
startTime = datetime.now()

# Use built in Nelder-Mead optimization algorithm.
# The calibration loop is implemented internally to the SciPy minimize function.
# The minimize function will call markov_model function every time it tries a new set of parameters.
# initial_params array contains the initial parameter quesses.
result = minimize(markov_model, initial_params, method='Nelder-Mead')

# Calculate the time it took to run the optimization loop.
endTime = datetime.now() - startTime

# Print the results.
print("Estimated Rate12  = %4.6f" % result.x[0])
print("Estimated Rate13  = %4.6f" % result.x[2])
print("Estimated Rate23  = %4.6f" % result.x[4])
print("Estimated Shape12 = %4.6f" % result.x[1])
print("Estimated Shape13 = %4.6f" % result.x[3])
print("Estimated Shape23 = %4.6f" % result.x[5])
print("Goodness of Fit value = %3.6e" % result.fun)

print("Analysis duration: %s sec." % endTime.total_seconds())
